__version__ = '0.2.2'
__author__ = "Diogo André"
__date__ = "2022-05-05"
__annotations__ = "Utility functions and data structure for Analog Integrated Circuit Modelling"
from .utils import *
from .read import *
from .write import *
from .data import *

def verbose_info():
    print(f"{__name__}")
    print(f"Version:        {__version__} ({__date__})")
    print(f"Author:         {__author__}")
    print(f"Description:    {__annotations__}")


from http.client import UnimplementedFileMode
from loguru import logger
import os
import yaml
import json
import toml
import csv
import pandas as pd
def write(info: dict, path:str) -> None:
    """_summary_
    Writes the contents of a dictionary to a YAML or JSON file
    containing the labelled information from within the dictionary
    Args:
        path (str): path to read the file from
    Raises:
        FileNotFoundError: _description_
        ValueError: _description_
        IOError: _description_
    """
    raise NotImplementedError(f"write function not implemented")
    if not os.path.exists(path):
        raise FileNotFoundError(f"File {path} not found")
    head,tail = os.path.split(path)
    name, extension = os.path.splitext(tail)
    possible_formats = [ ".yaml", ".yml", ".json", "toml"]
    try:
        with open(path, 'w') as file:
            if extension in [".yaml", ".yml"]:
                yaml.dump(info, file)
            elif extension == ".json":
                json.dump(info, file)
            elif extension == ".toml":
                toml.dump(info, file)
            else:
                raise ValueError(f"File {path} is not a valid specification file. Only {possible_formats} are accepted")
    except:
        raise IOError(f"File {path} could not be opened")

def write_data(data: pd.DataFrame, path: str):
    """_summary_
    Writes a CSV data file from a pandas dataframe
    Args:
        data (pd.DataFrame) : dataframe to write to the file
        path (str)          : path to read the file from

    Raises:
        FileNotFoundError: _description_
        ValueError: _description_
        IOError: _description_

    Returns:
        pandas DataFrame: dataframe containing the extracted information from the CSV file
    """
    raise NotImplementedError(f"write_data function not implemented")
    if not os.path.exists(path):
        raise FileNotFoundError(f"File {path} not found")
    head,tail = os.path.split(path)
    name, extension = os.path.splitext(tail)
    if extension !=  ".csv":
        raise ValueError(f"File {path} is not a valid specification file. Only .csv files are accepted")
    df = None
    try:
        data.to_csv(path)
    except:
        raise IOError(f"File {path} could not be opened")
    return df


""" ***********************************
* *[author] Diogo André (git-hub : das-dias)
* *[date] 2022-05-05
* *[filename] utils.py
* *[summary] Essential utilities for data processing and representation and other stuff
* ***********************************
"""
import pandas as pd
import itertools
from cycler import cycler
from enum import Enum
import numpy as np
from mpl_toolkits import mplot3d
import matplotlib as mpl
import matplotlib.pyplot as plt
import os
import seaborn as sns
__figs_path__ = os.path.join(os.getcwd(),"figs")
class Units(Enum):
    """_summary_
    Units enumerator
    """
    OHM="\u03A9"#capital Omega letter
    AMPERE="A"
    VOLTAGE="V"
    HERTZ="Hz"
    POWER="W"
    SIEMENS="S"
    FARAD="F"
    HENRY="H"
    METER="m"
    METER_SQR="m^2"
    DECIBEL="dB"
    TIME="s"
class Scale(Enum):
    """_summary_
    Unit scale enumerator
    """
    EXA=('E', pow(10,18))
    PETA=('P', pow(10,15))
    TERA=('T', pow(10,12))
    GIGA=('G', pow(10,9))
    MEGA=('M', pow(10,6))
    KILO=('k', pow(10,3))
    MILI=('m', pow(10,-3))
    MICRO=('u', pow(10,-6))
    NANO=('n', pow(10,-9))
    PICO=('p', pow(10,-12))
    FEMTO=('f', pow(10,-15))
    ATTO=('a', pow(10,-18))

def stof(val: str)->float:
    """_summary_
    Get a value from a string
    Args:
        val (str): String value in the format of "<value> <scale letter>"
    Raises:
        ValueError: invalid scaling factor
    Returns:
        float: the extracted floating point value
    """
    float_separators = [".", ","]
    scaling_factor = 1.0
    tokens = val.split(" ")
    res = 0.0
    for token in tokens:
        if token.isnumeric():
            res = float(token)
        else:
            if any([sep in token for sep in float_separators]):
                res = float(token)
            else:
                scale_letters = [scale.value[0] for scale in Scale]
                if token not in scale_letters:
                    raise ValueError(f"Invalid scaling factor: {token}")
                scaling_factor = [scale.value[1] for scale in Scale if scale.value[0] == token][0]
    res *= scaling_factor
    return res

def timer(func):
    """_summary_
    Decorator to time a function
    Args:
        func (function): function to be timed
    Returns:
        function: the same function, but with a "runtime_ns" (time in nanoseconds) attribute
    """
    import time
    from loguru import logger
    def wrapper(*args, **kwargs):
        start = time.time_ns()
        result = func(*args, **kwargs)
        end = time.time_ns()
        delta = end - start
        func.runtime_ns = delta
        delta_mili = (delta * 1e-6) # obtain time in miliseconds
        mili_secs = f"{Scale.MILI.value[0]}{Units.TIME.value}"
        logger.info(f"\nFunction: {func.__name__}\tRuntime: {delta_mili:.3f} {mili_secs}.")
        return result
    return wrapper

def _set_2D_style():
    sns.set_style('whitegrid') # darkgrid, white grid, dark, white and ticks
    custom_cycler = (
        cycler(color=['r', 'g', 'b', 'y']) +
        cycler(linestyle=['-', '--', ':', '-.'])
    )
    plt.rc("lines", linewidth=4)    # line width
    plt.rc('axes', titlesize=18)     # fontsize of the axes title
    plt.rc('axes', labelsize=14)    # fontsize of the x and y labels
    plt.rc("axes", prop_cycle=custom_cycler) # define a cycler of colours and linestyle
    plt.rc('xtick', labelsize=13)    # fontsize of the tick labels
    plt.rc('ytick', labelsize=13)    # fontsize of the tick labels
    plt.rc('legend', fontsize=13)    # legend fontsize
    plt.rc('font', size=13)          # controls default text sizes
    #define font family to use for all text
    mpl.rcParams['font.family'] = 'serif'

def _plot_graph_2D(
    x, 
    y, 
    label: str=None, 
    xlabel: str=None, 
    ylabel: str=None, 
    title: str=None, 
    filename: str=None,
    legend: bool=False,
    hold_off: bool=False,
    axis=None,
    show: bool=False,
    xlim: tuple=None,
    ylim: tuple=None
    ):
    if not bool(axis):
        axis = plt.axes()
    if not hold_off:
        axis.plot(x, y, label=label)  
        if bool(xlabel):
            axis.set_xlabel(xlabel)
        if bool(ylabel): 
            axis.set_ylabel(ylabel)
        if bool(title):    
            axis.set_title(title)
        if legend:
            axis.legend()
        if not bool(xlim):
            plt.xlim([x[0], x[-1]])
        else:
            plt.xlim([xlim[0], xlim[1]])
        if bool(ylim):
            plt.xlim([ylim[0], ylim[1]])
        if filename is not None:
            if not os.path.exists(__figs_path__):
                os.makedirs(__figs_path__)
            plt.savefig(os.path.join(__figs_path__,filename))
        if show:
            plt.show()
        plt.close()
    else:
        axis.plot(x, y, label=label)
    return axis
    
def _plot_graph_3D(
    x, 
    y, 
    z, 
    xlabel: str=None, 
    ylabel: str=None, 
    zlabel: str=None, 
    title: str=None, 
    filename: str=None, 
    type: str=None,
    show: bool=False
    ):
    # pretty plot
    fig = plt.figure()
    ax = plt.axes(projection='3d')
    type = type if bool(type) else "line"
    if type.lower() in ["line", None]:
        ax.plot3D(x,y,z)
    elif type.lower() == "scatter":
        c = x+y
        ax.scatter(x, y, z, c=c)
    elif type.lower() == "surface":
        ax.plot_surface(x, y, z, cmap="seismic", edgecolor="grey")
    elif type.lower() == "wireframe":
        ax.plot_wireframe(x, y, z, color='blue')
    else:
        raise ValueError(f"Unsupported plot type: {type}")
    
    if title is not None:
        ax.set_title(title)
    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    if zlabel is not None:
        ax.set_label(zlabel)
    if bool(filename):
        if not os.path.exists(__figs_path__):
            os.makedirs(__figs_path__)
        plt.savefig(os.path.join(__figs_path__,filename))
    if show:
        plt.show()
    plt.close()


def plot_function(
    x, 
    y,
    z = None, 
    labels: list=[], 
    xlabel: str=None, 
    ylabel: str=None,
    zlabel: str=None,
    title: str=None,
    filename: str=None,
    type: str=None,
    show: bool=False,
    xlim: tuple=None,
    ylim: tuple=None
    ):
    """_summary_
    Plots a function in 2D (curve) or 3D space (surface) depending on the parsing of the z variable
    Args:
        x           (np.ndarray / list) : x values
        y           (np.ndarray / list) : y values
        z           (np.ndarray / None) : z values
        labels      (list)              : list of labels for the plot
        xlabel      (str)               : abciss axis title
        ylabel      (str)               : ordinate axis title
        zlabel      (str)               : ordinate axis title
        title       (str)               : title of the plot
        filename    (str)               : name of the file to save the plot
        type        (str)               : type of 3D plot to be made. Options : "line", "scatter", "surface", "wireframe" 
        line_style  (str)               : line style to be used for the 2D plot
    """
    _set_2D_style()
    if not all([isinstance(label, str) for label in labels]):
        raise TypeError("Labels should be strings")
    if not isinstance(z, np.ndarray):
        if isinstance(y, np.ndarray):
            _plot_graph_2D(
                x, y,
                label = labels[0] if len(labels)>0 else None,
                xlabel=xlabel, 
                ylabel=ylabel, 
                title=title, 
                filename=filename,
                show=show,
                xlim=xlim,
                ylim=ylim
                )
        elif isinstance(y, list):
            if len(labels) not in [len(y), 0]:
                raise ValueError("Labels and y values should have the same length or no labels at all")
            if not all([isinstance(i, np.ndarray) for i in y]):
                raise ValueError("y must be a list of numpy.ndarray")
            if isinstance(x, list):
                if not all([isinstance(i, np.ndarray) for i in x]):
                    raise ValueError("x must be a list of numpy.ndarray")
                funcs = list(zip(x, y))
                axis = None
                for i, xy in enumerate(funcs[:-1]):
                    x_vec, y_vec = xy
                    axis = _plot_graph_2D(
                        x_vec, y_vec,
                        label = labels[i] if len(labels)>0 else None,
                        xlabel=xlabel, 
                        ylabel=ylabel, 
                        title=title, 
                        filename=filename, 
                        legend=False,
                        hold_off= True,
                        axis=axis
                    )
                x_vec, y_vec = funcs[-1]
                _plot_graph_2D(
                    x_vec, y_vec,
                    label = labels[-1] if len(labels)>0 else None,
                    xlabel=xlabel, 
                    ylabel=ylabel, 
                    title=title, 
                    filename=filename,
                    legend=True,
                    hold_off= False,
                    show=show,
                    axis=axis,
                    xlim=xlim,
                    ylim=ylim
                )
            elif isinstance(x, np.ndarray):
                axis = None
                for i, y_vec in enumerate(y[:-1]):
                    axis = _plot_graph_2D(
                        x, y_vec,
                        label = labels[i] if len(labels)>0 else None, 
                        xlabel=xlabel, 
                        ylabel=ylabel, 
                        title=title, 
                        filename=filename,
                        legend=False,
                        hold_off= True,
                        axis=axis
                    )
                y_vec = y[-1]
                _plot_graph_2D(
                    x, y_vec,
                    label = labels[-1] if len(labels)>0 else None,
                    xlabel=xlabel, 
                    ylabel=ylabel, 
                    title=title, 
                    filename=filename,
                    legend=True,
                    hold_off= False,
                    show=show,
                    axis=axis,
                    xlim=xlim,
                    ylim=ylim
                )
            else:
                raise ValueError("x must be a list of numpy.ndarray or a numpy.ndarray")
        else:
            raise ValueError("y must be a list of numpy.ndarray or a numpy.ndarray")
    else :
        if not all([isinstance(v, np.ndarray) for v in [x,y,z]]):
            raise TypeError("All vectors must be of type numpy.ndarray")
        _plot_graph_3D(
            x, y, z, 
            xlabel=xlabel, 
            ylabel=ylabel, 
            zlabel=zlabel, 
            title=title, 
            filename=filename, 
            type=type,
            show=show
        )


def plot_hist(data, labels: list=[], xlabel: str=None, title: str=None, filename: str=None, show:bool=False, stat: str="probability"):
    """_summary_
    Plots a histogram of the data
    Args:
        data        (np.ndarray / list) : data to be plotted 
        labels      (list, optional)    : Data labels. Defaults to [].
        title       (str, optional)     : Title of the histogram plot. Defaults to None.
        filename    (str, optional)     : Name of the figure to save the histogram. Defaults to None.
    Raises:
        TypeError: _description_
        ValueError: _description_
        ValueError: _description_
        TypeError: _description_
    """
    _set_2D_style()
    if not all([isinstance(label, str) for label in labels]):
        raise TypeError("Labels should be strings")
    
    if isinstance(data, np.ndarray):
        sns.histplot(
            data=data,
            kde=True,
            binwidth = 0.05,
            stat=stat,
        ).set_xlabel(xlabel)
    elif isinstance(data, list):
        if len(labels) not in [len(data), 0]:
            raise ValueError("Labels and data must have the same length")
        
        if not all([isinstance(v, np.ndarray) for v in data]):
            raise ValueError("data must be a list of numpy.ndarray")
        dt = data
        if len(labels) != 0:
            dt = {label: v for label, v in zip(labels, data)}
        colours = [
            "#000000",
            "#D84242",
            "#4131F0",
            "#FF0000",
            "#D80AD0"
        ]
        sns.histplot( 
            data=dt, 
            kde=True, 
            binwidth=0.05,
            legend=True, 
            stat=stat,
            palette=[c for c,_ in zip(itertools.cycle(colours), dt.keys())]
        ).set_xlabel(xlabel)
    else:
        raise TypeError("data must be a numpy.ndarray or a list of numpy.ndarray")
    if bool(title):
        plt.title(title)
    if bool(filename):
        if not os.path.exists(__figs_path__):
            os.makedirs(__figs_path__)
        plt.savefig(os.path.join(__figs_path__,filename))
    if show:
        plt.show()
    plt.close()
    
    


from collections import defaultdict
from itertools import cycle
from loguru import logger
import os
import toml
import json
import csv
import pandas as pd
from copy import copy

from .utils import(
    Scale,
    stof
)
from .data import(
    Devices,
)

def read_specs(path:str) -> Devices:
    """_summary_
    Reads the contents of YAML and JSON files and returns a dictionary
    containing the labelled information from within the files
    Args:
        path (str): path to read the file from

    Raises:
        FileNotFoundError: _description_
        ValueError: _description_
        IOError: _description_

    Returns:
        dict: data structure containing the extracted information from the YAML / JSON file
    """
    if not os.path.exists(path):
        raise FileNotFoundError(f"File {path} not found")
    head,tail = os.path.split(path)
    name, extension = os.path.splitext(tail)
    if extension not in [ ".toml", ".json"]:
        raise ValueError(f"File {path} is not a valid specification file. Only .toml and .json are accepted")
    struct = {}
    try:
        with open(path, 'r') as file:
            if extension == ".toml":
                struct = toml.load(file)
            elif extension == ".json":
                struct = json.load(file)
    except:
        raise IOError(f"File {path} could not be read")
    devices = Devices()
    devices.parse_data(struct)
    return devices if bool(struct) else None

def read_data(path: str) -> pd.DataFrame:
    """_summary_
    Reads a CSV data file and returns a pandas dataframe
    Args:
        path (str): path to read the file from

    Raises:
        FileNotFoundError: _description_
        ValueError: _description_
        IOError: _description_

    Returns:
        pandas DataFrame: dataframe containing the extracted information from the CSV file
    """
    if not os.path.exists(path):
        raise FileNotFoundError(f"File {path} not found")
    head,tail = os.path.split(path)
    name, extension = os.path.splitext(tail)
    if extension !=  ".csv":
        raise ValueError(f"File {path} is not a valid specification file. Only .csv files are accepted")
    df = None
    try:
        with open(path, 'r') as file:
            df = pd.read_csv(file)
    except:
        raise IOError(f"File {path} could not be read")
    return df

def read_lut(path: str) -> pd.DataFrame:
    """_summary_
    Reads a Cadence Look Up Table exported to CSV
    and unfolds it to return a Pandas DataFrame that only
    includes raw axis
    Args:
        path (str): path to read the file from

    Raises:
        FileNotFoundError: _description_
        ValueError: _description_
        IOError: _description_

    Returns:
        pandas DataFrame: dataframe containing the extracted information from the CSV file
    """
    lut = read_data(path)
    original_lut_size = len(lut)
    # detect variables present in lut name
    head,tail = os.path.split(path)
    name, extension = os.path.splitext(tail)
    attrs = name.split('_')
    detected_vars={}
    sweeped_vars = []
    for attr in attrs:
        tokens=attr.split('-')
        var_name = tokens[0]
        if var_name == "sweep":
            sweeped_vars.append([tokens[1], tokens[2]])
            break
        var_alpha=""
        scale = None
        for token in tokens[1:]:
            if token.isnumeric():
                var_alpha = ".".join([var_alpha,token]) if var_alpha != "" else token
            else:
                scale = token
        #convert the detected variable into float
        # and multiply it by the respective found scale
        var = 0.0
        try:
            var = float(var_alpha)
        except:
            raise ValueError(f"Wrong LUT naming format: {name}.")
        scaling_factor = 1.0
        if scale != None:
            scaling_letters = [s.value[0] for s in Scale]
            if scale not in scaling_letters:
                raise ValueError(f"Unrecognized unit scaling token: {scale}")
            scaling_factor = [s.value[1] for s in Scale if s.value[0] == scale][0]
        var = var*scaling_factor
        # add the var and var_name to the detected_vars dict
        detected_vars[var_name] = var
# detect the variable sweeps present in the name of each column
    data=defaultdict(list)
    sweep_axis_value_space = defaultdict(list)
    for column in lut.columns[1:]:
        tokens = column.split(' ')
        var_name = tokens[0].split(':')[1]
        [data[var_name].append(val) for val in lut[column].values]
        if len(tokens)>2:
            var_name = tokens[1]
            var_value = float(tokens[2])
            if bool(sweep_axis_value_space.get(var_name)):
                if var_value not in sweep_axis_value_space.get(var_name):
                    sweep_axis_value_space[var_name].append(var_value)
            else:
                sweep_axis_value_space[var_name].append(var_value)
    max_col_len = 0
    # get the maximum true length of the entire expanded lut table
    for col in data.keys():
        if len(data[col]) > max_col_len:
            max_col_len = len(data[col])
    # adjoint the constant axis : vsb and lx
    for var_name in detected_vars.keys():
        data[var_name] = [detected_vars[var_name]]*max_col_len
    # adjoint the secondary sweeping variable - vds or vsd
    # and expand the short axis (x-axis) until it reaches the required length
    # adjoining the primary sweeping axis onto the data frame
    x_axis = lut.columns[0].split(' ')[0]
    x_axis = x_axis.replace(' ','')
    for var_name in sweep_axis_value_space.keys():
        for var_value in sweep_axis_value_space[var_name]:
            [data[var_name].append(val) for val in [var_value]*original_lut_size]
            [data[x_axis].append(val) for val in lut[lut.columns[0]].values]
    return pd.DataFrame(data)


""" ***********************************
* *[author] Diogo André (git-hub : das-dias)
* *[date] 21-05-2022
* *[filename] data.py
* *[summary] Python lang file implementing the data structures enabling the
*               device control variables parsing environment
* ***********************************
"""
from collections import defaultdict
from re import S
import numpy as np
from pandas import DataFrame
from enum import Enum
import os
from .utils import(
    Scale,
    Units,
    stof
)


class TomlSections(Enum):
    """_summary_
    Enumerator for the types of sections in
    parsing toml files
    Args:
        Enum (_type_): _description_
    """
    CONTROL="control"
    SPIT="spit"

class TomlControlKeywords(Enum):
    """_summary_
    Enumerator for the possible keywords
    in the parsing toml file's [control] section
    Args:
        Enum (_type_): _description_
    """
    DEVICE="device"
    DEVICES="devices"
    VARACTOR="varactor"
    VARACTORS="varactors"
    TYPE="type"
    CVAR="cvar" # target varactor capacitance
    VGB="vgb"
    VBG="vbg"
    VDS="vds"
    VSD="vsd"
    GMOVERID="gmoverid"
    LENGTH="l"
    VSB="vsb"
    VBS="vbs"
    ID="id"

class TomlSpitKeywords(Enum):
    """_summary_
    Enumerator for the possible keywords
    in the parsing toml file's [spit] section
    Args:
        Enum (_type_): _description_
    """
    OUTPUTDIR="outputdir"
    VARS="vars"
    PLOT="plot"

class TomlSpitVariables(Enum):
    """_summary_
    Enumerator for the possible keywords
    in the parsing toml file's [spit] posssible variables
    Args:
        Enum (_type_): _description_
    """
    VGS="vgs"
    VSG="vsg"
    WIDTH="w"
    GDS="gds"
    GM="gm"
    GMBS="gmbs"
    SELF_GAIN="selfgain"
    F_T="ft"
    F_OSC="fosc"
    CGS="cgs"
    CGD="cgd"
    CGB="cgb"
    CSB="csb"
    CDB="cdb"
    CVAR="cvar" # actual varactor capacitance
    R_ON="ron"
    REGION="region"
    ALL="all"

class TomlSpitPlot(Enum):
    TRUE="true"
    FALSE="false"
    
class TomlControlType(Enum):
    NMOS="nch"
    PMOS="pch"
    
class MosCell:
    """_summary_
    Object implementing the CMOS device
    control and main spice parameters
    Args:
        object (_type_): _description_
    """
    __slots__=[
        "name",
        "type",
        "vgb",
        "vgb",
        "vbg",
        "vds",
        "vsd",
        "vsb",
        "vbs",
        "gmoverid",
        "l",
        "id",
        "vgs",
        "vsg",
        "w",
        "cgs",
        "cgd",
        "cgb",
        "csb",
        "cdb",
        "cvar",
        "gm",
        "gmbs",
        "gds",
        "self_gain",
        "ft",
        "fosc",
        "ron",
        "region",
        "vdsat",
    ]
    __UNITS__={
        "name":"",
        "type":"",
        "vgb":Units.VOLTAGE.value,
        "vbg":Units.VOLTAGE.value,
        "vds":Units.VOLTAGE.value,
        "vsd":Units.VOLTAGE.value,
        "vsb":Units.VOLTAGE.value,
        "vbs":Units.VOLTAGE.value,
        "gmoverid":Units.VOLTAGE.value+"^-1",
        "l":Units.METER.value,
        "id":Units.AMPERE.value,
        "vgs":Units.VOLTAGE.value,
        "vsg":Units.VOLTAGE.value,
        "w":Units.METER.value,
        "cgs":Units.FARAD.value,
        "cgd":Units.FARAD.value,
        "cgb":Units.FARAD.value,
        "csb":Units.FARAD.value,
        "cdb":Units.FARAD.value,
        "cvar":Units.FARAD.value,
        "gmbs":Units.SIEMENS.value,
        "gm":Units.SIEMENS.value,
        "gds":Units.SIEMENS.value,
        "self_gain":"VV^-01",
        "ft":Units.HERTZ.value,
        "fosc":Units.HERTZ.value,
        "ron":Units.OHM.value,
        "region":"",
        "vdsat":Units.VOLTAGE.value,
    }
    def __init__(
        self,
        name:str="m0",
        type:str="nch",
        vds:float=0.0,
        vsb:float=0.0,
        gmoverid:float=1.0,
        l:float=30*Scale.NANO.value[1],
        id:float=None,
        vgb:float=0.0,
        cvar:float =0.0
    ):
        self.name=name
        self.type=type
        # input vars
        # varactor 
        self.vgb = vgb if bool(vgb) else None
        self.vbg = -self.vgb if bool(self.vgb) else None
        self.cvar=cvar
        # cell
        self.vds:float=vds
        self.vsd:float=-vds
        self.vsb:float=vsb
        self.vbs:float=-vsb
        self.gmoverid:float=gmoverid
        self.l:float=l
        self.id:float=id
        # output vars
        self.vgs:float=None
        self.vsg:float=-self.vgs if bool(self.vgs) else None
        self.w:float=None
        self.cgs:float=None
        self.cgd:float=None
        self.cgb:float=None
        self.csb:float=None
        self.cdb:float=None
        self.gmbs:float=None
        self.gm:float=None
        self.gds:float=None
        self.self_gain:float=None
        self.ft:float=None
        self.fosc:float=None
        self.ron:float=None
        self.region:float=None
        self.vdsat:float=None
        
        
    def __str__(self)->str:
        vars = [var for var in dir(self) if not var.startswith('__')]
        vals = [getattr(self, var) for var in vars]
        obj={key:val for (key,val) in zip(vars,vals)}
        result=""
        for var,val in obj.items():
            if type(val)==float:
                result+="{:.2f} ".format(val)
            else:
                result += f"{val} "
        return result

    def __parse_data__(self, key:str, val)->None:
        if key == TomlControlKeywords.DEVICE.value or key == TomlControlKeywords.VARACTOR.value:
            if type(val) != str:
                raise ValueError("Device name must be a string")
            self.name = val
        elif key == TomlControlKeywords.TYPE.value:
            if type(val) != str:
                raise ValueError("Device type must be a string")
            if val not in [TomlControlType.NMOS.value, TomlControlType.PMOS.value]:
                raise ValueError(f"Device type must be \"{TomlControlType.NMOS.value}\" or \"{TomlControlType.PMOS.value}\"")
            self.type = val
        elif key == TomlControlKeywords.CVAR.value:
            if type(val) not in [float, str, int]:
                raise ValueError("Varactor Cap. CVAR must be parsed as a float, integer or string")
            if type(val) == str:
                self.cvar = stof(val)
            else:
                self.cvar = float(val)
        elif key == TomlControlKeywords.VGB.value:
            if self.type != TomlControlType.NMOS.value:
                raise ValueError("VGB value to be specified is only valid for NMOS devices")
            if type(val) not in [float, str, int]:
                raise ValueError("VGB value must parsed as a float, integer or string")
            if type(val) == str:
                self.vgb = stof(val)
                self.vbg = -self.vgb
            else:
                self.vgb = float(val)
                self.vbg = -self.vgb
        elif key == TomlControlKeywords.VBG.value:
            if self.type != TomlControlType.PMOS.value:
                raise ValueError("VBG value to be specified is only valid for PMOS devices")
            if type(val) not in [float, str, int]:
                raise ValueError("VBG value must parsed as a float, integer or string")
            if type(val) == str:
                self.vbg = stof(val)
                self.vgb = -self.vbg
            else:
                self.vbg = float(val)
                self.vgb = -self.vbg
        elif key == TomlControlKeywords.VDS.value:
            if self.type != TomlControlType.NMOS.value:
                raise ValueError("VDS value to be specified is only valid for NMOS devices")
            if type(val) not in [float, str, int]:
                raise ValueError("VDS value must parsed as a float, integer or string")
            if type(val) == str:
                self.vds = stof(val)
                self.vsd = -self.vds
            else:
                self.vds = float(val)
                self.vsd = -self.vds
        elif key == TomlControlKeywords.VSD.value:
            if self.type != TomlControlType.PMOS.value:
                raise ValueError("VSD value to be specified is only valid for PMOS devices")
            if type(val) not in [float, str, int]:
                raise ValueError("VSD value must parsed as a float, integer or string")
            if type(val) == str:
                self.vsd = stof(val)
                self.vds = -self.vsd
            else:
                self.vsd = float(val)
                self.vds = -self.vsd
        elif key == TomlControlKeywords.VSB.value:
            if self.type != TomlControlType.NMOS.value:
                raise ValueError("VSB value to be specified is only valid for NMOS devices")
            if type(val) not in [float, str, int]:
                raise ValueError("Vsb value must parsed as a float, integer or string")
            if type(val) == str:
                self.vsb = stof(val)
                self.vbs = -self.vsb
            else:
                self.vsb = float(val)
                self.vbs = -self.vsb
        elif key == TomlControlKeywords.VBS.value:
            if self.type != TomlControlType.PMOS.value:
                raise ValueError("VBS value to be specified is only valid for PMOS devices")
            if type(val) not in [float, str, int]:
                raise ValueError("VBS value must parsed as a float, integer or string")
            if type(val) == str:
                self.vbs = stof(val)
                self.vsb = -self.vbs
            else:
                self.vbs = float(val)
                self.vsb = -self.vsb
        elif key == TomlControlKeywords.GMOVERID.value:
            if type(val) not in [float, str, int]:
                raise ValueError("Gm/Id value must parsed as a float, integer or string")
            if type(val) == str:
                self.gmoverid = stof(val)
            else:
                self.gmoverid = float(val)
        elif key == TomlControlKeywords.LENGTH.value:
            if type(val) not in [float, str]:
                raise ValueError("Channel Length (L) value must parsed as a float or string")
            if type(val) == str:
                self.l = stof(val)
            else:
                self.l = val
        elif key == TomlControlKeywords.ID.value:
            if type(val) not in [float, str]:
                raise ValueError("Drive Current (Id) value must parsed as a float or string")
            if type(val) == str:
                self.id = stof(val)
            else:
                self.id = val
        else:
            raise ValueError(f"{key} is not a valid control variable for a CMOS device")
                
        
class Devices:
    """_summary_
    A class object ot save all the devices to which
    the gm/id sizing process will be applied, and the output
    specifications of each device
    """
    __slots__=["devices", "varactors", "spits", "plot", "output_dir"]
    def __init__(self):
        self.devices={}
        self.varactors={}
        self.spits=defaultdict(list)
        self.plot=False
        self.output_dir=None
    
    def __data_frame__(self, cvar=False)->DataFrame:
        """_summary_
        Converts the device dictionary to a Pandas dataframe
        Returns:
            DataFrame: _description_
        """
        vars = [var for var in dir(MosCell) if not var.startswith('__')]
        columns = [var+"["+MosCell.__UNITS__[var]+"]" for var in vars]
        data = {k:[] for k in columns}
        index = None
        if not cvar:
            for device in self.devices.values():
                for i,var in enumerate(vars):
                    data[columns[i]].append(getattr(device, var))
            order = ["name", "type", "vds", "vsb", "gmoverid", "l", "w", "id"]
            order = [var+"["+MosCell.__UNITS__[var]+"]" for var in order]
            order = order+[var for var in columns if var not in order]
            index = list(self.devices.keys())
        else:
            for varactor in self.varactors.values():
                for i,var in enumerate(vars):
                    data[columns[i]].append(getattr(varactor, var))
            order = ["name", "type", "vgb", "l", "w", "cvar"]
            order = [var+"["+MosCell.__UNITS__[var]+"]" for var in order]
            order = order+[var for var in columns if var not in order]
            index = self.varactors.keys()
        return DataFrame(data, index=index)[order]

    def __str__(self) -> str:
        devices_df = self.__data_frame__()
        vars_df = self.__data_frame__(True)
        ret = "Devices:\n"
        ret += str(devices_df) if not devices_df.empty else "--None--\n"
        ret += "Varactors:\n"
        ret += str(vars_df) if not vars_df.empty else "--None--\n"
        return ret
    
    def add(self, device, varactor = False) -> None:
        if varactor:
            self.varactors[device.name] = device
        else:
            self.devices[device.name] = device

    def parse_data(self, data:dict) -> None:
        """_summary_
        Parses the data dictionary and adds the devices to the Devices object
        Args:
            data (dict): Dictionary containing the data to be parsed
        """
        for sec in data.keys():
            if sec == TomlSections.CONTROL.value:
                # get first data input of the control section
                key,token = list(data[sec].items())[0]
                if key == TomlControlKeywords.DEVICE.value:
                    # create a new device
                    device = MosCell(name=token)
                    for key,token in list(data[sec].items())[1:]:
                        device.__parse_data__(key, token)
                    self.add(device)
                elif key == TomlControlKeywords.VARACTOR.value:
                    # create a new varactor
                    varactor = MosCell(name=token)
                    for key,token in list(data[sec].items())[1:]:
                        varactor.__parse_data__(key, token)
                    self.add(varactor, True)
                elif key == TomlControlKeywords.DEVICES.value:
                    devices_names = [n for n in token]
                    for key,token in list(data[sec].items())[1:]:
                        if key not in devices_names:
                            raise ValueError(f"{key} is an unrecognized device name")
                        device = MosCell(name=key)
                        for subkey, subtoken in token.items():
                            device.__parse_data__(subkey, subtoken)
                        self.add(device)
                elif key == TomlControlKeywords.VARACTORS.value:
                    varactors_names = [n for n in token]
                    for key,token in list(data[sec].items())[1:]:
                        if key not in varactors_names:
                            raise ValueError(f"{key} is an unrecognized device name")
                        varactor = MosCell(name=key)
                        for subkey, subtoken in token.items():
                            varactor.__parse_data__(subkey, subtoken)
                        self.add(varactor, True)
                else:
                    raise ValueError(f"{key} is not a valid key for the first argument of {TomlSections.CONTROL.name} section. First argument must be {TomlControlKeywords.DEVICE.name} or {TomlControlKeywords.DEVICES.name}.")
            elif sec == TomlSections.SPIT.value:
                for key,token in data[sec].items():
                    if key == TomlSpitKeywords.OUTPUTDIR.value:
                        if type(token) != str:
                            raise ValueError(f"{TomlSpitKeywords.OUTPUTDIR.value} must be a string")
                        if not os.path.isdir(token):
                            os.mkdir(token)
                        self.output_dir = token
                    elif key == TomlSpitKeywords.PLOT.value:
                        if type(token) not in [bool, str]:
                            raise ValueError(f"{TomlSpitKeywords.PLOT.value} must be a boolean or a string")
                        if type(token) == str:
                            self.plot = token.lower() == "true"
                        else:
                            self.plot = token
                    elif key == TomlSpitKeywords.VARS.value:
                        acceptable_vars = ['all']
                        [acceptable_vars.append(v) for v in dir(MosCell) if not v.startswith('__')]
                        for subkey, subtoken in token.items():
                            
                            if subkey not in list(self.devices.keys()) + list(self.varactors.keys()):
                                raise ValueError(f"{subkey} is an unrecognized device name. Recognized devices are {list(self.devices.keys()) + list(self.varactors.keys())}")
                            for var in subtoken:    
                                if var not in acceptable_vars:
                                    raise ValueError(f"{var} is an unrecognized variable name")
                                if var != 'all':
                                    self.spits[subkey].append(var)
            else:
                raise ValueError(f"{sec} is not a valid section name")
        if len(self.devices) == 0 and len(self.varactors) == 0:
            raise ValueError("No devices were detected in the TOML setup file")


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['modelling_utils']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'cycler>=0.11.0,<0.12.0',
 'loguru>=0.6.0,<0.7.0',
 'matplotlib>=3.5.2,<4.0.0',
 'numpy>=1.22.3,<2.0.0',
 'pandas>=1.4.2,<2.0.0',
 'seaborn>=0.11.2,<0.12.0',
 'toml>=0.10.2,<0.11.0']

setup_kwargs = {
    'name': 'modelling-utils',
    'version': '0.2.2',
    'description': ' Utility functions and data structure for aiding in Analog Integrated Circuit Modelling using Python',
    'long_description': None,
    'author': 'Diogo André',
    'author_email': 'das.dias6@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/das-dias/modelling_utils',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)
